/*
 * USB HC08 driver 
 * based on the usb-skeleton driver
 * and the usb-tower driver
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <linux/completion.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/devfs_fs_kernel.h>

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/signal.h>
#include <linux/poll.h>
#include <linux/fcntl.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/devfs_fs_kernel.h>


#include <linux/moduleparam.h>

#define DEBUG

/* Use our own dbg macro 
 * and enable debug by default */
#undef dbg
#ifdef DEBUG	
	static int debug = 1;	
#define dbg(format, arg...) printk(KERN_DEBUG  __FILE__": " format "\n", ## arg)
#else	
	static int debug = 0;	
#define dbg(format, arg...) 
#endif

/* Version Information */
#define DRIVER_VERSION "0.2"
#define DRIVER_AUTHOR "simon"
#define DRIVER_DESC "USB HC08 Driver"

module_param(debug, int, 0);
//MODULE_PARAM_DESC(debug, "debug enabled or not");

#define USB_HC08_VENDOR_ID	0x0c70
#define USB_HC08_PRODUCT_ID	0x0000

/* table of devices that work with this driver. 
 * The usb system use this table to match our device
 * and then call the probe function */
static struct usb_device_id hc08_table [] = {
	{ USB_DEVICE(USB_HC08_VENDOR_ID, USB_HC08_PRODUCT_ID) },
	{ }					/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, hc08_table);

/* Get a minor range for your devices from the usb maintainer */
#define USB_HC08_MINOR_BASE	192	

/* number of devices allowed to be plug in the same time 
 * arbitrary number... can be change depending needs */
#define MAX_DEVICES		2

#define URB_TIMEOUT (2*HZ) /* timeout between a urb submit 
			      and a effective  transfert*/


/* nb of buffers we can stock by advance */
#define NB_BUF	8

/* Structure to hold all of our device specific stuff */
struct usb_hc08 {
	struct usb_device *	udev;			/* save off the usb device pointer */
	struct usb_interface *	interface;		/* the interface for this device */
	unsigned char		minor;			/* the starting minor number for this device */
	unsigned char		num_ports;		/* the number of ports this device has */
	char			num_interrupt_in;	/* number of interrupt in endpoints we have */
	char			num_int_in;		/* number of bulk in endpoints we have */
	char			num_int_out;		/* number of bulk out endpoints we have */

	unsigned char *		read_buffer;		/* circular buffer for the user */
	unsigned char *		read_buffer_start;	
	int			read_buffer_size;	
	int			read_buffer_max_size;	

	unsigned char *		int_in_buffer;		/* the buffer to receive data */
	int			int_in_size;		/* the size of the receive buffer */
	struct urb *		read_urb;
	__u8			int_in_endpointAddr;	/* the address of the bulk in endpoint */
	__u8			int_in_interval;	/* the interval in ms for each transfert */
	wait_queue_head_t	int_in_wait;		
	
	unsigned char *		int_out_buffer;		/* the buffer to send data */
	int			int_out_size;		/* the size of the send buffer */
	struct urb *		write_urb;		/* the urb used to send data */
	__u8			int_out_endpointAddr;	/* the address of the bulk out endpoint */
	__u8			int_out_interval;	/* the interval in ms for each transfert */
	wait_queue_head_t	int_out_wait;		

	int			open_count;		/* number of times this port has been opened */
	struct semaphore	sem;			/* locks this structure */
};


/* local function prototypes */
static ssize_t hc08_read	(struct file *file, char *buffer, size_t count, loff_t *ppos);
static ssize_t hc08_write	(struct file *file, const char *buffer, size_t count, loff_t *ppos);
static int hc08_ioctl		(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int hc08_open		(struct inode *inode, struct file *file);
static int hc08_release		(struct inode *inode, struct file *file);
static int hc08_probe	(struct usb_interface *interface, const struct usb_device_id *id);
static void hc08_disconnect	(struct usb_interface *interface);
static void hc08_write_int_callback (struct urb *urb, struct pt_regs *regs);
static void hc08_read_int_callback (struct urb *urb,struct pt_regs *regs);
static void usb_hc08_debug_data (const char *, int, const unsigned char *);
static void hc08_delete (struct usb_hc08 *);
static void hc08_unlink_urb (struct usb_hc08 *);

/*
 * File operations needed when we register this driver.
 * This assumes that this driver NEEDS file operations,
 * of course, which means that the driver is expected
 * to have a node in the /dev directory. 
 */
static struct file_operations hc08_fops = {
	/*
	 * The owner field is part of the module-locking
	 * mechanism. The idea is that the kernel knows
	 * which module to increment the use-counter of
	 * BEFORE it calls the device's open() function.
	 * This also means that the kernel can decrement
	 * the use-counter again before calling release()
	 * or should the open() function fail.
	 *
	 * Not all device structures have an "owner" field
	 * yet. "struct file_operations" and "struct net_device"
	 * do, while "struct tty_driver" does not. If the struct
	 * has an "owner" field, then initialize it to the value
	 * THIS_MODULE and the kernel will handle all module
	 * locking for you automatically. Otherwise, you must
	 * increment the use-counter in the open() function
	 * and decrement it again in the release() function
	 * yourself.
	 */
	owner:		THIS_MODULE,

	read:		hc08_read,
	write:		hc08_write,
	ioctl:		hc08_ioctl,
	open:		hc08_open,
	release:	hc08_release,
};      


/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver hc08_driver = {
	owner:		THIS_MODULE,
	name:		"hc08",
	probe:		hc08_probe,
	disconnect:	hc08_disconnect,
	id_table:	hc08_table,
};


static struct usb_class_driver hc08_class = {
	name:		"usb/hc08_%d",
	fops:		&hc08_fops,
	mode:		S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP,
	minor_base:	USB_HC08_MINOR_BASE,
};

/*
 *	usb_hc08_debug_data
 */
static void usb_hc08_debug_data (const char *function, int size, const unsigned char *data)
{
	int i;

	if (!debug)
		return;
	
	printk (KERN_DEBUG __FILE__": %s - length = %d, data = ", 
		function, size);
	
	for (i = 0; i < size; ++i)
	{
		printk ("%.2x ", data[i]);
	}
	printk ("\n");
}

/*
 *	hc08_delete
 */
static void hc08_delete (struct usb_hc08 *dev)
{
	dbg("enter %s", __FUNCTION__);
	if (dev->write_urb != NULL)
		usb_free_urb (dev->write_urb);
	if (dev->read_urb != NULL)
		usb_free_urb (dev->read_urb);
	if (dev->int_in_buffer != NULL)
		kfree (dev->int_in_buffer);
	if (dev->int_out_buffer != NULL)
		kfree (dev->int_out_buffer);
	kfree (dev);
}

/*
 *	hc08_unlink_urb
 */
static void hc08_unlink_urb (struct usb_hc08 *dev)
{
	        usb_unlink_urb (dev->write_urb);
		if (debug)
		{
			if (dev->write_urb->status != -ECONNRESET)
				dbg("%s - fail to unlink write urb ( status : %d )", 
						__FUNCTION__, dev->write_urb->status);
		}
		usb_unlink_urb (dev->read_urb);
		if (debug)
		{
			if (dev->read_urb->status != -ECONNRESET)
				dbg("%s - fail to unlink read urb ( status : %d )", 
						__FUNCTION__, dev->read_urb->status);
		}
}

/*
 *	hc08_open
 */
static int hc08_open (struct inode *inode, struct file *file)
{
	struct usb_hc08 *dev = NULL;
	struct usb_interface *interface;
	int subminor;
	int retval = -ENODEV;
	
	dbg ("enter %s", __FUNCTION__);
	
	subminor = iminor (inode); 

	interface = usb_find_interface (&hc08_driver, subminor);
	if (!interface)
	{
		err ("%s:  - error, can't find device #%d", 
				__FUNCTION__, 
				USB_HC08_MINOR_BASE - subminor);
		goto exit;
		
	}
	
	info ("%s:  - device #%d found",
			__FUNCTION__, 
			USB_HC08_MINOR_BASE - subminor);
		
	dev = usb_get_intfdata (interface);
	if (!dev)
		goto exit;

	/* lock this device */
	down (&dev->sem);

	/* increment our usage count for the driver */
	++dev->open_count;
	
	/* save our object in the file's private structure */
	file->private_data = dev;
	
	retval = 0;
	
	up (&dev->sem);

exit :
	return retval;
}

/*
 *	hc08_release
 */
static int hc08_release (struct inode *inode, struct file *file)
{
	struct usb_hc08 *dev;
	int retval = 0;

	dbg ("enter %s", __FUNCTION__);
	
	dev = (struct usb_hc08 *)file->private_data;
	
	if (dev == NULL) 
	{
		dbg ("%s - object is NULL", __FUNCTION__);
		return -ENODEV;
	}

	dbg(" - device %d", USB_HC08_MINOR_BASE - dev->minor);

	/* lock our device */
	down (&dev->sem);

	if (dev->open_count <= 0)
	{
		dbg ("%s - device not opened", __FUNCTION__);
		retval = -ENODEV;
		goto exit_not_opened;
	}

	if (dev->udev == NULL) 
	/* the device was unplugged before the file was released */
	{
		dbg("%s - device unplugged", __FUNCTION__);
		hc08_unlink_urb (dev);
		up (&dev->sem);
		hc08_delete (dev);
		return 0;
	}

	/* decrement our usage count for the device */
	--dev->open_count;
	if (dev->open_count <= 0) 
	{
		mdelay (1000);
		hc08_unlink_urb (dev);
		dev->open_count = 0;
	}

exit_not_opened:
	up (&dev->sem);

	return retval;
}


/*
 *	hc08_read
 */
static ssize_t hc08_read (struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	int retval = 0, timeout = URB_TIMEOUT;
	struct usb_hc08 *dev;
	size_t bytes_reads = 0;

	dev = (struct usb_hc08 *)file->private_data;
	
	dbg ("enter %s: - device #%d, count = %d", __FUNCTION__, 
			USB_HC08_MINOR_BASE - dev->minor, count);

	/* verify that the device wasn't unplugged */
	if (dev->udev == NULL) 
	{
		retval = -ENODEV;
		goto exit;
	}
	
	/* if a transfer is already occuring... we sleep, 
	 * waiting for the transfert end */
/*	while (dev->read_urb->status == -EINPROGRESS)
	{
		retval = wait_event_interruptible (dev->int_in_wait, 
				dev->read_urb->status);
		if (retval < 0)
		{	
			err ("%s - error wait_event_interruptible () ",
					__FUNCTION__);
			goto exit;
		}
	}
*/	
	/* lock the device structure */
	down (&dev->sem);
	
	/* if there is no enough data in the buffer... 
	 * we try to get more...  */
	//if (dev->read_buffer_size < count)
	if (1)
	{	
		/* set our read urb */
		usb_fill_int_urb (dev->read_urb, 
				 dev->udev, 
				 usb_rcvintpipe (dev->udev, 
					 dev->int_in_endpointAddr), 
				 dev->int_in_buffer, 
				 dev->int_in_size, 
				 hc08_read_int_callback, dev, dev->int_in_interval);

				up (&dev->sem);
				
				/* submit our urb each time we want a read */
				retval = usb_submit_urb (dev->read_urb, GFP_KERNEL);
				
				//up (&dev->sem);
				
				/* if a transfer is already occuring... we sleep, 
				 * waiting for the transfert end */
	
				while (dev->read_urb->status == -EINPROGRESS)
				{
					retval = wait_event_interruptible (dev->int_in_wait, 
							dev->read_urb->status);
					if (retval < 0)
					{	
						err ("%s - error wait_event_interruptible () ", __FUNCTION__);
						goto exit;
					}
				} 
				
				/* we sleep... waiting for data */
		/*		retval = interruptible_sleep_on_timeout (&dev->int_in_wait, 
				timeout);
				if (retval < 0)
				{
					err ("%s: - error interruptible_sleep_on_timeout () ", 
							__FUNCTION__);
					goto exit;
				}	*/
	}
	dbg ("%s: read_buffer_size : %d", __FUNCTION__,
			dev->read_buffer_size);

	/*  we can only read as much the buffer can contain */
	bytes_reads = (count > dev->read_buffer_size) ? 
				dev->read_buffer_size : count;

	down (&dev->sem);

	/* if we have some data for the user...*/
//	if (bytes_reads)
//	{
		/* read and copy the data to userspace */
		if (copy_to_user (buffer, dev->read_buffer_start, bytes_reads))
		{	
			retval = -EFAULT;
		}	
		else
		{	
			dev->read_buffer_start += bytes_reads;
			dev->read_buffer_size -= bytes_reads;
			retval = bytes_reads;
		}	
//	}
	
exit :
	up (&dev->sem);
	return retval;
}


/*
 *	hc08_write
 */
static ssize_t hc08_write (struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
	struct usb_hc08 *dev;
	ssize_t bytes_written = 0;
	int retval = 0;

	dev = (struct usb_hc08 *)file->private_data;

	dbg("%s: - device #%d, count = %d", __FUNCTION__
			, USB_HC08_MINOR_BASE - dev->minor, count);

	/* lock this object */
	down (&dev->sem);

	/* verify that the device wasn't unplugged */
	if (dev->udev == NULL) {
		retval = -ENODEV;
		goto exit;
	}

	/* verify that we actually have some data to write */
	if (count == 0) 
	{
		err (" - write request of 0 bytes");
		goto exit;
	}

	/* we can only write as much as 1 urb will hold */
	bytes_written = (count > dev->int_out_size) ? 
				dev->int_out_size : count;
	
	/* copy the data from userspace into our urb */
	if (copy_from_user(dev->int_out_buffer, buffer, bytes_written)) 
	{
		retval = -EFAULT;
		goto exit;
	}
	
	/* set up our write urb */
	usb_fill_int_urb (dev->write_urb, 
			  dev->udev, 
		          usb_sndintpipe (dev->udev, dev->int_out_endpointAddr),
		          dev->int_out_buffer, 
			  dev->int_out_size,
		          hc08_write_int_callback, 
			  dev, dev->int_out_interval);
	
	/* if debug = 1 , more informations are given about the buffer content*/
	usb_hc08_debug_data (__FUNCTION__, bytes_written, 
			     dev->int_out_buffer);
	
/*	while (dev->write_urb->status == -EINPROGRESS)
	{
		retval = wait_event_interruptible (dev->int_out_wait, 
				dev->write_urb->status);
		if (retval < 0)
		{	
			err ("%s - wait_event_interruptible () ", __FUNCTION__);
			goto exit;
		}
	}
*/	
	retval = usb_submit_urb (dev->write_urb, GFP_KERNEL);

	if (retval) 
	{
		err("%s - failed submitting write urb, error %d",
				__FUNCTION__, retval);
	}
	else 
	{
		retval = bytes_written;
	}
	
	/* if a transfer is already occuring... we sleep, 
	 * waiting for the transfert end */
	
	while (dev->write_urb->status == -EINPROGRESS)
	{
		retval = wait_event_interruptible (dev->int_out_wait, 
				dev->write_urb->status);
		if (retval < 0)
		{	
			err ("%s - error wait_event_interruptible () ", __FUNCTION__);
			goto exit;
		}
	}
	
	retval = bytes_written;
	
exit:
	/* unlock the device */
	up (&dev->sem);

	return retval;
}


/*
 *	hc08_ioctl ??
 */
static int hc08_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct usb_hc08 *dev;
	int retval = -ENOTTY;

	dev = (struct usb_hc08 *)file->private_data;

	/* lock this object */
	down (&dev->sem);

	/* verify that the device wasn't unplugged */
	if (dev->udev == NULL) {
		up (&dev->sem);
		return -ENODEV;
	}

	dbg(" - device #%d, cmd 0x%.4x, arg %ld", 
	    USB_HC08_MINOR_BASE - dev->minor, cmd, arg);


	switch (cmd)
	{
		/* FIXME */
	}
	
	/* unlock the device */
	up (&dev->sem);
	
	/* return that we did not understand this ioctl call */
	return retval;
}

/*
 *	hc08_write_int_callback
 */
static void hc08_write_int_callback (struct urb *urb, struct pt_regs *regs)
{
	struct usb_hc08 *dev = (struct usb_hc08 *)urb->context;

	dbg("%s: - device #%d", __FUNCTION__, 
			USB_HC08_MINOR_BASE - dev->minor);

	if ((urb->status == -ENOENT) || 
	    (urb->status == -ECONNRESET)) 
	{
		err ("%s - write urb cancelled assynchronously or by user ", __FUNCTION__);
	}	
	else
	{	
		err ("%s - write int status : %d", __FUNCTION__, urb->status);
	}
	
	/* send a signal to the sleeping process,
	 * waiting for write again */
	
	wake_up_interruptible (&dev->int_out_wait);

	return;
}

/*
 *	hc08_read_int_callback
 *	this function is call each time a read transfert is polled
 *	valid data are put into the circular buffer
 */
static void hc08_read_int_callback (struct urb *urb, struct pt_regs *regs)
{
	struct usb_hc08 *dev = (struct usb_hc08 *)urb->context;

	dbg("%s: - device #%d", __FUNCTION__,
			USB_HC08_MINOR_BASE - dev->minor);
	
	if ((urb->status == -ENOENT) ||
	    (urb->status == -ECONNRESET)) 
	{
		err ("%s - read urb cancelled assynchronously or by user ", __FUNCTION__);
	}
	else
	{	
		dbg ("%s - read int status : %d", __FUNCTION__, urb->status);
	}
	
	down (&dev->sem);

	/* we are only interessed by new data */
	if (urb->actual_length > 0)
	{	
		// print via debug the recept data 
		usb_hc08_debug_data (__FUNCTION__, dev->int_in_size, 
				dev->int_in_buffer);
		/* we add the recept data to the circular buffer 
		 *
		 *        _________________________________________________________________________________
		 *        #		      o	o o o o o o o o o o					  #
		 *        #		      o	o o o o o o o o o o					  #
		 *        #		      o	o o  DATA   o o o o					  #
		 *        #		      o	o o o o o o o o o o					  #
		 *        #		      o	o o o o o o o o o o					  #
		 *        #_________+_________+_________+_________+_________+_________+_________+_________#
		 *
		 *    read_buffer            / \                 / \                                
		 *       / \		      |                   |                                      / \
		 *        |          read_buffer_start	          |			                  |
		 *        |      	      |			  |			                  |
		 *	  |	              |-------------------|      			          |
		 *	  |	                read_buffer_size   					  |
		 *	  |            							                  |
		 *	  |-------------------------------------------------------------------------------|
		 *	         read_buffer_max_size ( actualy by default = 8 bytes )
		 */
		
			if (dev->read_buffer_start + dev->read_buffer_size + urb->actual_length < 
					dev->read_buffer + dev->read_buffer_max_size)
			{
			// if we have enought place to put the data 
				memcpy (dev->read_buffer_start + dev->read_buffer_size,
						dev->int_in_buffer, urb->actual_length);
				dev->read_buffer_size += urb->actual_length;
			}
			else
			{
			// else we put their at the begining	
				dev->read_buffer_size = (urb->actual_length < dev->read_buffer_max_size) 
						? urb->actual_length : dev->read_buffer_max_size;
				memcpy (dev->read_buffer, dev->int_in_buffer, 
						dev->read_buffer_size);
				dev->read_buffer_start = dev->read_buffer;
			}	
	}

	dbg ("%s: read_buffer_size : %d", __FUNCTION__, dev->read_buffer_size);

	/* send a signal to the sleeping process,
	* waiting for reading */
	wake_up_interruptible (&dev->int_in_wait);
	
	up (&dev->sem);
	
	return;
}

/*
 *	hc08_probe
 *
 *	Called by the usb core when a new device is connected that it thinks
 *	this driver might be interested in.
 *	if it is the case, we process at a lot of initalisations
 */
static int hc08_probe (struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(interface);
	struct usb_host_interface *iface_desc;
	struct usb_hc08 *dev = NULL;
	struct usb_endpoint_descriptor *endpoint;
	int buffer_size;
	int num_endpoints;
	int i, endpoint_match = 0, retval = -ENOMEM;

	
	/* See if the device offered us matches what we can accept */
	if ((udev->descriptor.idVendor != USB_HC08_VENDOR_ID) ||
	    (udev->descriptor.idProduct != USB_HC08_PRODUCT_ID)) 
	{
		info ("unable to probe new device");
		retval = -ENODEV;
		goto exit;
	}

	/* allocate memory for our device state and intialize it */
	dev = (struct usb_hc08 *) kmalloc (sizeof(struct usb_hc08), GFP_KERNEL);
	if (dev == NULL) {
		err ("Out of memory");
		goto exit;
	}
	memset (dev, 0x00, sizeof (*dev));

	init_MUTEX (&dev->sem);
	dev->udev = udev;
	dev->interface = interface;

	/* wait queue initialisation */

	init_waitqueue_head (&dev->int_out_wait);
	init_waitqueue_head (&dev->int_in_wait);

	/* set up the endpoint information */
	/* check out the endpoints */
	iface_desc = &interface->altsetting[0];
	num_endpoints = iface_desc->desc.bNumEndpoints;
	
	for (i = 0; i < num_endpoints; ++i) 
	{
		endpoint = &iface_desc->endpoint[i].desc;
		
		if ((endpoint->bmAttributes & 0) == 0x03)
			dbg ("maybe control endpoint found");

		if ((endpoint->bEndpointAddress & USB_DIR_IN) && 
		   ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) 
		    == USB_ENDPOINT_XFER_INT)) 
		{
			/* we found a interrupt in endpoint */
			dev->read_urb = usb_alloc_urb(0, GFP_KERNEL);

			if (!dev->read_urb) 
			{
				err("No free urbs available");
				goto error;
			} 
			buffer_size = endpoint->wMaxPacketSize;
			dev->int_in_size = buffer_size;
			dev->int_in_endpointAddr = endpoint->bEndpointAddress;
			dev->int_in_interval = endpoint->bInterval;
			dev->int_in_buffer = kmalloc (buffer_size, GFP_KERNEL);
			if (!dev->int_in_buffer) 
			{
				err("Couldn't allocate int_in_buffer");
				goto error;
			}
			/* we can stock NB_BUF buffer for the user */
			dev->read_buffer_max_size = NB_BUF * buffer_size;
			dev->read_buffer = kmalloc (dev->read_buffer_size, GFP_KERNEL);
			if (!dev->read_buffer) 
			{
				err("Couldn't allocate int_in_buffer");
				goto error;
			}	
			dev->read_buffer_size = 0;
			dev->read_buffer_start = dev->read_buffer;
			endpoint_match ++;

			/* initialisation of the interrupt read buffer */
			memset (dev->int_in_buffer, 0x55, dev->int_in_size);
		}
	
		
		if (!(endpoint->bEndpointAddress & USB_DIR_IN) && 
		   ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) 
		    == USB_ENDPOINT_XFER_INT)) 
		{
			/* we found a interrupt out endpoint */
			dev->write_urb = usb_alloc_urb(0, GFP_KERNEL);

			if (!dev->write_urb) 
			{
				err("No free urbs available");
				goto error;
			} 
			buffer_size = endpoint->wMaxPacketSize;
			dev->int_out_size = buffer_size;
			dev->int_out_endpointAddr = endpoint->bEndpointAddress;
			/* as we want write data packet one by one 
			 * the interval use to poll urb transfert is
			 * fix to 0 
			 * that's mean... only one transfert */
			dev->int_out_interval = endpoint->bInterval;
			dev->int_out_buffer = kmalloc (buffer_size, GFP_KERNEL);
			if (!dev->int_out_buffer) 
			{
				err("Couldn't allocate int_out_buffer");
				goto error;
			}
		
			endpoint_match ++;
		}	
	}

	dbg (" : number of endpoint match %d", endpoint_match);
	/* if no endpoint is matching, just exit*/
	if (endpoint_match < 2)
	{
		info ("only %d enpoints found insteed of 2",
				endpoint_match);
		retval = -ENODEV;
		goto error;
	}	

	/* register the device now */
	usb_set_intfdata (interface, dev);
	retval = 0; 

	retval = usb_register_dev (interface, &hc08_class);
	if (retval) 
	{
		/* something prevented us from registering this driver */
		err ("Not able to get a minor for this device.");
		usb_set_intfdata (interface, NULL);
		goto error;
	} 
			
	dev->minor = interface->minor;

	/* let the user know what node this device is now attached to */
	info ("USB hc08 driver now attached to hc08_%d", 
			USB_HC08_MINOR_BASE - dev->minor);
	goto exit;
	
error:
	hc08_delete (dev);

exit:
	return retval;
}


/*
 *	hc08_disconnect
 *
 *	Called by the usb core when the device is removed from the system.
 */
static void hc08_disconnect(struct usb_interface *interface)
{
	struct usb_hc08 *dev;
	int minor;

	dbg("enter %s", __FUNCTION__);
	
	dev = usb_get_intfdata (interface);
	usb_set_intfdata (interface, NULL);
	down (&dev->sem);
		
	minor = dev->minor;

	if (!dev->open_count) 
	{
		dbg("%s : device not open ", __FUNCTION__);
		up (&dev->sem);
		hc08_delete (dev);
	} 
	else 	
	/* the release function will call hc08_delete () ... later */
	{
		dbg("%s : device open ???? ", __FUNCTION__);
		dev->udev = NULL;
		up (&dev->sem);
	}

	info("USB hc08 #%d now disconnected", 
			minor - USB_HC08_MINOR_BASE);
}



/*
 *	usb_hc08_init
 */
static int __init usb_hc08_init(void)
{
	int result;

	/* register this driver with the USB subsystem 
	 * see the usb_driver structure for more details */
	result = usb_register(&hc08_driver);
	if (result < 0) {
		err ("usb_register failed. Error number %d",
		    result);
		return -1;
	}
	
	dbg("%s: init complete", __FUNCTION__);
	info(DRIVER_DESC "" DRIVER_VERSION);
	return 0;
}


/*
 *	usb_hc08_exit
 */
static void __exit usb_hc08_exit(void)
{
	dbg("entrer %s", __FUNCTION__);

	/* deregister this driver with the USB subsystem */
	usb_deregister(&hc08_driver);
}


module_init (usb_hc08_init);
module_exit (usb_hc08_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
